"""Main code for jodo."""
import argparse
import os
import sys
from typing import Iterator, Optional

from pycloudlib import (
    EC2,
    GCE,
    OCI,
    Azure,
    LXDContainer,
    LXDVirtualMachine,
    Openstack,
    Qemu,
)
from pycloudlib.cloud import BaseCloud
from pycloudlib.instance import BaseInstance
from pycloudlib.util import log_exception_list

from jodo import db

CLOUDS: dict[str, type[BaseCloud]] = {
    "azure": Azure,
    "ec2": EC2,
    "gce": GCE,
    "lxd_container": LXDContainer,
    "lxd_vm": LXDVirtualMachine,
    "oci": OCI,
    "openstack": Openstack,
    "qemu": Qemu,
}


def launch(
    cloud: str,
    name: str,
    image: str,
    instance_type: Optional[str] = None,
    userdata: Optional[str] = None,
) -> None:
    """Launch a new instance."""
    cloud_class = CLOUDS[cloud](tag=name, timestamp_suffix=False)
    try:
        image_id = cloud_class.released_image(image)
    except Exception:
        image_id = image
    kwargs = {"image_id": image_id}
    if instance_type:
        kwargs["instance_type"] = instance_type
    if userdata:
        kwargs["user_data"] = userdata
    db.ensure_not_exists(name=name)
    instance: BaseInstance = cloud_class.launch(**kwargs)
    instance.wait()
    db.create_info(instance, cloud=cloud, name=name)


def list_instances() -> Iterator[str]:
    """List instances."""
    results = db.list_instances()
    table_fields = [x[1:] for x in results]
    table = [("NAME", "CLOUD", "IP", "PORT", "CREATED"), *table_fields]

    # Calculate the max width of each column then add 2 for more spacing
    widths = [max(len(v) for v in col) + 2 for col in zip(*table, strict=True)]

    # Now print the table
    for row in table:
        yield "".join(f"{row[index]:<{widths[index]}}" for index in range(len(row)))


def execute(name: str, command: str, sudo: bool = False) -> str:
    """Execute a command on an instance."""
    result = _get_instance(name).execute(command, use_sudo=sudo)
    output = f"return code: {result.return_code}\nstdout:\n{result.stdout}\n"
    if result.stderr:
        output += f"stderr:\n{result.stderr}"
    return output


def push(name: str, source: str, destination: str) -> None:
    """Push a file to an instance."""
    _get_instance(name).push_file(local_path=source, remote_path=destination)


def pull(name: str, source: str, destination: str) -> None:
    """Pull a file from an instance."""
    _get_instance(name).pull_file(remote_path=source, local_path=destination)


def ssh(name: str) -> None:
    """SSH to an instance."""
    info = db.get_info(name)
    os.execlp(  # noqa: S606
        "ssh",
        "ssh",
        f"ubuntu@{info.ip}",
        "-p",
        info.port,
    )


def delete(name: str) -> None:
    """Delete an instance."""
    instance = _get_instance(name)
    exceptions = instance.delete()
    db.delete_info(name)
    if exceptions:
        print(
            "Deleted instance info, but got problems deleting on cloud. "
            "Ensure instance is actually deleted."
        )
    log_exception_list(exceptions)


def _get_instance(name: str) -> BaseInstance:
    """Get an instance."""
    try:
        instance_info = db.get_info(name)
    except ValueError:
        print(f"Could not find instance '{name}'. Does it exist?")
        sys.exit(1)
    cloud = CLOUDS[instance_info.cloud](tag=instance_info.name)
    return cloud.get_instance(instance_info.instance_id)


def take_action(args: argparse.Namespace) -> None:
    """Take appropriate action based on arguments."""
    match args.command:
        case "launch":
            launch(
                args.cloud,
                args.name,
                args.image,
                args.instance_type,
                args.userdata,
            )
        case "list":
            print("\n".join(list_instances()))
        case "ssh":
            ssh(args.name)
        case "delete":
            delete(args.name)
        case "exec":
            print(execute(args.name, args.exec_command))
        case "push":
            push(args.name, args.source, args.destination)
        case "pull":
            pull(args.name, args.source, args.destination)
        case _:
            raise ValueError("Argparse shouldn't let us get here")  # noqa: TRY003


def parse_args() -> argparse.Namespace:
    """Parse and return arguments."""
    parser = argparse.ArgumentParser(description="One CLI to rule them all")
    subparsers = parser.add_subparsers(help="sub-command", dest="command")

    parser_launch = subparsers.add_parser("launch", help="Launch a new instance")
    parser_launch.add_argument("name")
    parser_launch.add_argument("-c", "--cloud", choices=CLOUDS.keys())
    parser_launch.add_argument("-i", "--image", required=True)
    parser_launch.add_argument("-t", "--instance_type")
    parser_launch.add_argument("-u", "--userdata")

    subparsers.add_parser("list", help="List instances")

    parser_ssh = subparsers.add_parser("ssh", help="SSH to an instance")
    parser_ssh.add_argument("name")

    parser_delete = subparsers.add_parser("delete", help="delete an instance")
    parser_delete.add_argument("name", nargs="?", default=None)

    parser_execute = subparsers.add_parser(
        "exec",
        help="Execute a command on an instance",
    )
    parser_execute.add_argument("-s", "--sudo", action="store_true")
    parser_execute.add_argument("name")
    parser_execute.add_argument("exec_command")

    parser_push = subparsers.add_parser("push", help="Push a file to an instance")
    parser_push.add_argument("name")
    parser_push.add_argument("source")
    parser_push.add_argument("destination")

    parser_pull = subparsers.add_parser("pull", help="Pull a file from an instance")
    parser_pull.add_argument("name")
    parser_pull.add_argument("source")
    parser_pull.add_argument("destination")

    return parser.parse_args()


def main() -> None:
    """Run main function."""
    take_action(parse_args())


if __name__ == "__main__":
    main()



"""Module for models and database related operations."""
import sqlite3
from contextlib import suppress
from pathlib import Path
from typing import NamedTuple

from pycloudlib.instance import BaseInstance

SCHEMA_VERSION = 1
DB_DIR = Path.home() / ".config" / "jodo"


class InstanceExistsError(Exception):
    """Raised when instance already exists in database."""

    def __init__(self, name: str):
        """Initialize AlreadyExistsError."""
        super().__init__(f"Instance with name '{name}' already exists.")


class InstanceNotFoundError(Exception):
    """Raised when instance cannot be found in database."""

    def __init__(self, name: str):
        """Initialize InstanceNotFoundError."""
        super().__init__(f"Instance '{name}' cannot be found.")


class InstanceInfo(NamedTuple):
    """Instance details."""

    instance_id: str
    name: str
    cloud: str
    ip: str
    port: str
    created: str


def initialize_db(
    db_dir: Path = DB_DIR,
) -> sqlite3.Connection:
    """Initialize database."""
    db_dir.mkdir(parents=True, exist_ok=True)
    db_path = db_dir / "jodo.db"

    connection = sqlite3.connect(str(db_path), isolation_level=None)
    connection.execute(f"PRAGMA user_version = {SCHEMA_VERSION}")

    connection.execute(
        """
        CREATE TABLE IF NOT EXISTS instance (
            id TEXT PRIMARY KEY,
            name TEXT,
            cloud TEXT,
            ip TEXT,
            port TEXT,
            created TEXT
        )
    """,
    )
    return connection


connection: sqlite3.Connection = initialize_db()


def ensure_not_exists(name: str) -> None:
    with suppress(InstanceNotFoundError):
        get_info(name)
        raise InstanceExistsError(name)


def create_info(instance: BaseInstance, cloud: str, name: str) -> None:
    """Add newly launched instance to database."""
    ensure_not_exists(name)
    connection.execute(
        "INSERT INTO instance (id, name, cloud, ip, port, created) VALUES "
        "(?, ?, ?, ?, ?, datetime('now')) ",
        (str(instance.id), name, cloud, instance.ip, instance.port),
    )


def list_instances() -> list[InstanceInfo]:
    """List all instances in database."""
    cursor = connection.execute(
        "SELECT id, name, cloud, ip, port, created FROM instance",
    )
    return [InstanceInfo(*i) for i in cursor.fetchall()]


def get_info(name: str) -> InstanceInfo:
    """Get instance from database."""
    result: list[tuple] = connection.execute(
        "SELECT id, name, cloud, ip, port, created FROM instance WHERE name=?",
        (name,),
    ).fetchall()
    rowcount = len(result)
    if rowcount != 1:
        raise InstanceNotFoundError(name)
    return InstanceInfo(*result[0])


def delete_info(name: str) -> None:
    """Delete instance from database."""
    result = connection.execute("DELETE FROM instance WHERE name = ?", (name,))
    if result.rowcount == 0:
        InstanceNotFoundError(name)
    connection.commit()



import re
import subprocess
from pathlib import Path

import pytest

from jodo import main
from jodo.db import InstanceExistsError


class TestMain:
    """Test jodo/main.py."""

    name = "jodo-test-container"

    @pytest.fixture(scope="class")
    def _instance(self):
        """Create an instance for use in other tests."""
        assert "jodo_test_container" not in subprocess.check_output(
            ["lxc", "list"],
            universal_newlines=True,
        )
        main.launch(
            cloud="lxd_container",
            name=self.name,
            image="jammy",
            userdata="#cloud-config\nbootcmd:\n - echo 'hi' > /var/tmp/hi",
        )
        yield
        main.delete(self.name)

    def test_no_duplicates(self, _instance) -> None:
        """Test no duplicates."""
        with pytest.raises(InstanceExistsError):
            main.launch(cloud="lxd_container", name=self.name, image="jammy")
        assert "\n".join(main.list_instances()).count(self.name) == 1

    def test_list(self, _instance) -> None:
        """Test list_instances."""
        instances = "\n".join(main.list_instances())
        assert re.search(rf"{self.name}\s*lxd_container", instances)

    def test_execute(self, _instance) -> None:
        """Test execute."""
        assert "hi" in main.execute(self.name, command="cat /var/tmp/hi")

    def test_push(self, _instance, tmp_path: Path) -> None:
        """Test push."""
        local_tmp = tmp_path / "push.txt"
        local_tmp.write_text("push data")
        main.push(self.name, source=str(local_tmp), destination="/tmp/push.txt")
        assert "push data" in main.execute(self.name, "cat /tmp/push.txt")

    def test_pull(self, _instance, tmp_path: Path) -> None:
        """Test pull"""
        local_tmp = tmp_path / "pull.txt"
        main.execute(self.name, "echo 'pull data' > /tmp/pull.txt")
        main.pull(self.name, "/tmp/pull.txt", str(local_tmp))
        assert "pull data" in local_tmp.read_text()
